"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("./operand");
var Hook_1 = require("./hooks/Hook");
var Label_1 = require("./Label");
var Compilation_1 = require("./Compilation");
var Asm = (function () {
    function Asm(opts) {
        var _this = this;
        this.expressions = [];
        this.hooks = {
            mnemonic: new Hook_1.default(['mnemonic', 'operands', 'opts']),
            command: new Hook_1.default(['name', 'args']),
            op: new Hook_1.default(['operand']),
            ops: new Hook_1.default(['operands', 'size']),
            instruction: new Hook_1.default([]),
            compilation: new Hook_1.default(['compilation']),
        };
        this._ = function (name) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            return _this.command.apply(_this, [name].concat(args));
        };
        if (typeof opts !== 'object' || !opts)
            throw new TypeError('Need to provide options object to Asm.');
        this.insert(this.lbl(opts.main || 'main'));
        if (typeof opts.littleEndian === void 0)
            opts.littleEndian = true;
        if (typeof opts.operandSize === void 0)
            opts.operandSize = operand_1.SIZE.D;
        if (typeof opts.addressSize === void 0)
            opts.addressSize = operand_1.SIZE.D;
        var plugins = opts.plugins;
        for (var _i = 0, plugins_1 = plugins; _i < plugins_1.length; _i++) {
            var plugin = plugins_1[_i];
            plugin.asm = this;
            plugin.onAsm(this);
        }
        this.opts = opts;
        this._.lbl = this.lbl.bind(this);
        this._.op = this.op.bind(this);
        this._.ops = this.ops.bind(this);
        this._.insert = this.insert.bind(this);
    }
    Asm.prototype.command = function (name) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        return this.hooks.command.call(name, args);
    };
    Asm.prototype.code = function (template) {
        template(this._);
    };
    Asm.prototype.replace = function (expr, index) {
        if (index === void 0) { index = this.expressions.length; }
        expr.index = index;
        expr.asm = this;
        this.expressions[index] = expr;
        expr.calcOffsetMaxAndOffset();
        return expr;
    };
    Asm.prototype.insert = function (expr) {
        this.replace(expr, this.expressions.length);
        expr.build();
        return expr;
    };
    Object.defineProperty(Asm.prototype, "label", {
        get: function () {
            return this.expressions[0];
        },
        enumerable: true,
        configurable: true
    });
    Asm.prototype.op = function (str) {
        return this.hooks.op.call(str);
    };
    Asm.prototype.ops = function (operands, size) {
        if (size === void 0) { size = this.opts.operandSize; }
        return this.hooks.ops.call(operands, size);
    };
    Asm.prototype.lbl = function (name) {
        return new Label_1.default(name);
    };
    Asm.prototype.instruction = function () {
        return this.hooks.instruction.call();
    };
    Asm.prototype.compile = function (container) {
        var compilation = new Compilation_1.default(this);
        this.hooks.compilation.call(compilation);
        return compilation.compile(container);
    };
    Asm.prototype.toString = function (lineNumbers, hex) {
        if (lineNumbers === void 0) { lineNumbers = true; }
        if (hex === void 0) { hex = true; }
        var lines = [];
        for (var i = 0; i < this.expressions.length; i++) {
            var expr = this.expressions[i];
            var lineNum = '';
            if (lineNumbers) {
                lineNum = i + '';
                if (lineNum.length < 3)
                    lineNum = ((new Array(4 - lineNum.length)).join('0')) + lineNum;
                lineNum += ' ';
            }
            lines.push(lineNum + expr.toString('    ', hex));
        }
        return lines.join('\n');
    };
    return Asm;
}());
exports.default = Asm;
