"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var expression_1 = require("./expression");
var CodeBuffer_1 = require("./CodeBuffer");
var Hook_1 = require("./hooks/Hook");
var Compilation = (function () {
    function Compilation(asm) {
        this.bin = null;
        this.hooks = {
            compiled: new Hook_1.default([]),
        };
        this.asm = asm;
    }
    Compilation.prototype.compile = function (container) {
        this.do2ndPass();
        var bin = this.bin = this.createBuffer();
        this.do3rdPass(bin);
        this.hooks.compiled.call();
        if (!container)
            return bin.truncate();
        else {
            for (var i = 0; i < bin.cursor; i++)
                container.push(bin.buf[i]);
            return container;
        }
    };
    Compilation.prototype.doPass = function (callback) {
        var expressions = this.asm.expressions;
        for (var i = 0; i < expressions.length; i++)
            callback(expressions[i], i);
    };
    Compilation.prototype.createBuffer = function () {
        var lastExpression = this.asm.expressions[this.asm.expressions.length - 1];
        var maxLength = lastExpression.offsetMax + lastExpression.bytesMax();
        return new CodeBuffer_1.default(maxLength);
    };
    Compilation.prototype.do2ndPass = function () {
        var _this = this;
        this.doPass(function (expression, index) {
            if (expression instanceof expression_1.ExpressionVolatile) {
                var fixedExpression = expression.getFixedSizeExpression();
                _this.asm.replace(fixedExpression, index);
                fixedExpression.calcOffset();
            }
            else {
                expression.calcOffset();
            }
        });
    };
    Compilation.prototype.do3rdPass = function (buffer) {
        this.doPass(function (expression, index) {
            if (expression instanceof expression_1.ExpressionVariable)
                expression.evaluate();
            expression.write(buffer);
        });
    };
    return Compilation;
}());
exports.default = Compilation;
