"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("./operand");
var Mnemonic = (function () {
    function Mnemonic(mnemonic, opcode) {
        if (mnemonic === void 0) { mnemonic = ''; }
        if (opcode === void 0) { opcode = 0x00; }
        this.operandSize = operand_1.SIZE.NONE;
        this.mnemonic = mnemonic;
        this.opcode = opcode;
    }
    Mnemonic.toStringOperand = function (operand) {
        var typeofOperand = typeof operand;
        if (typeofOperand === 'number')
            return operand;
        if (typeofOperand === 'string')
            return operand;
        if ((typeofOperand === 'object') || (typeofOperand === 'function')) {
            if (typeof operand.name === 'string')
                return operand.name;
            if (typeof operand.atomName === 'string')
                return operand.atomName;
            if (typeof operand.toString === 'function')
                return operand.toString();
        }
        return String(operand);
    };
    Mnemonic.toStringOperandShort = function (operand) {
        var typeofOperand = typeof operand;
        if (typeofOperand === 'number')
            return operand;
        if (typeofOperand === 'string')
            return operand;
        if ((typeofOperand === 'object') || (typeofOperand === 'function')) {
            if (typeof operand.atomName === 'string')
                return operand.atomName;
            if (typeof operand.name === 'string')
                return operand.name;
            if (typeof operand.toString === 'function')
                return operand.toString();
        }
        return String(operand);
    };
    Mnemonic.prototype.getName = function () {
        var size = this.operandSize;
        if ((size === operand_1.SIZE.ANY) || (size === operand_1.SIZE.NONE))
            return this.mnemonic;
        return this.mnemonic + operand_1.SIZE[size].toLowerCase();
    };
    Mnemonic.prototype.toJsonOperands = function () {
        var ops = [];
        for (var _i = 0, _a = this.operandTemplates; _i < _a.length; _i++) {
            var operandTemplate = _a[_i];
            var arr = [];
            for (var _b = 0, operandTemplate_1 = operandTemplate; _b < operandTemplate_1.length; _b++) {
                var op = operandTemplate_1[_b];
                arr.push(Mnemonic.toStringOperand(op));
            }
            if (arr.length > 1)
                ops.push(arr);
            else
                ops.push(arr[0]);
        }
        return ops;
    };
    Mnemonic.prototype.toJson = function () {
        var json = {
            mnemonic: this.mnemonic,
            mnemonicPrecise: this.getName(),
            opcode: this.opcode,
            opcodeHex: this.opcode.toString(16).toUpperCase(),
            operands: this.toJsonOperands(),
        };
        if (this.operandSize)
            json.operandSize = this.operandSize;
        return json;
    };
    Mnemonic.prototype.toString = function () {
        var opcode = ' ' + (new operand_1.Constant(this.opcode, false)).toString();
        var operands = [];
        for (var _i = 0, _a = this.operandTemplates; _i < _a.length; _i++) {
            var ops = _a[_i];
            var opsarr = ops.map(function (op) { return Mnemonic.toStringOperandShort(op); });
            operands.push(opsarr.join('/'));
        }
        var operandsstr = '';
        if (operands.length)
            operandsstr = ' ' + operands.join(', ');
        var size = '';
        if (this.operandSize > 0)
            size = ' ' + this.operandSize + '-bit';
        return this.mnemonic + size + opcode + operandsstr;
    };
    return Mnemonic;
}());
exports.Mnemonic = Mnemonic;
exports.default = Mnemonic;
