"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("./operand");
var util_1 = require("./util");
exports.SIZE_UNKNOWN = -Infinity;
exports.OFFSET_UNKNOWN = -Infinity;
var Expression = (function () {
    function Expression() {
        this.index = 0;
        this.length = exports.SIZE_UNKNOWN;
        this.offset = exports.OFFSET_UNKNOWN;
        this.offsetMax = exports.OFFSET_UNKNOWN;
        this.asm = null;
    }
    Expression.prototype.bytes = function () {
        return this.length;
    };
    Expression.prototype.bytesMax = function () {
        return this.bytes();
    };
    Expression.prototype.isFixedSize = function () {
        if (this.length === exports.SIZE_UNKNOWN)
            return false;
        return this.bytes() === this.bytesMax();
    };
    Expression.prototype.hasSize = function () {
        return this.bytes() !== exports.SIZE_UNKNOWN;
    };
    Expression.prototype.build = function () {
    };
    Expression.prototype.calcOffsetMaxAndOffset = function () {
        if (this.index === 0) {
            this.offsetMax = 0;
            this.offset = 0;
        }
        else {
            var prev = this.asm.expressions[this.index - 1];
            if (prev.offsetMax === exports.OFFSET_UNKNOWN)
                this.offsetMax = exports.OFFSET_UNKNOWN;
            else {
                var bytes_max = prev.bytesMax();
                if (bytes_max === exports.SIZE_UNKNOWN)
                    this.offsetMax = exports.OFFSET_UNKNOWN;
                else
                    this.offsetMax = prev.offsetMax + bytes_max;
            }
            if (prev.offset === exports.OFFSET_UNKNOWN)
                this.offset = exports.OFFSET_UNKNOWN;
            else {
                if (!prev.isFixedSize())
                    this.offset = exports.OFFSET_UNKNOWN;
                else
                    this.offset = prev.offset + prev.bytes();
            }
        }
    };
    Expression.prototype.calcOffset = function () {
        if (this.offset !== exports.OFFSET_UNKNOWN)
            return;
        if (this.index === 0)
            this.offset = 0;
        else {
            var prev = this.asm.expressions[this.index - 1];
            var offset = prev.offset;
            if (offset === exports.OFFSET_UNKNOWN)
                throw Error("Instruction [" + (this.index - 1) + "] does not have offset.");
            else {
                var bytes = prev.bytes();
                if (bytes === exports.SIZE_UNKNOWN)
                    throw Error("Instruction [" + (this.index - 1) + "] does not have size.");
                else
                    this.offset = offset + bytes;
            }
        }
    };
    Expression.prototype.rel = function (offset) {
        if (offset === void 0) { offset = 0; }
        return new operand_1.Relative(this, offset);
    };
    Expression.prototype.write = function (arr) { };
    Expression.prototype.toNumber = function () {
        return this.offset;
    };
    Expression.prototype.formatOffset = function () {
        var offset = '______';
        if (this.offset >= 0) {
            offset = this.offset.toString(16).toUpperCase();
            offset = (new Array(7 - offset.length)).join('0') + offset;
        }
        var max_offset = '______';
        if (this.offsetMax >= 0) {
            max_offset = this.offsetMax.toString(16).toUpperCase();
            max_offset = (new Array(7 - max_offset.length)).join('0') + max_offset;
        }
        return offset + '|' + max_offset;
    };
    Expression.prototype.formatToString = function (margin, expression, comment) {
        if (comment === void 0) { comment = ''; }
        expression = margin + expression;
        var spaces = util_1.repeat(' ', Math.max(0, Expression.commentColls - expression.length));
        return "" + expression + spaces + "; " + this.formatOffset() + " " + comment;
    };
    Expression.prototype.toString = function (margin, comment) {
        if (margin === void 0) { margin = ''; }
        if (comment === void 0) { comment = true; }
        return this.formatToString(margin, '[expression]');
    };
    Expression.commentColls = 44;
    return Expression;
}());
exports.Expression = Expression;
var ExpressionVariable = (function (_super) {
    __extends(ExpressionVariable, _super);
    function ExpressionVariable(ops, opts) {
        if (ops === void 0) { ops = null; }
        var _this = _super.call(this) || this;
        _this.ops = null;
        _this.isEvaluated = false;
        _this.ops = ops;
        _this.opts = opts;
        return _this;
    }
    ExpressionVariable.prototype.evaluate = function () {
        this.isEvaluated = true;
        return true;
    };
    return ExpressionVariable;
}(Expression));
exports.ExpressionVariable = ExpressionVariable;
var ExpressionVolatile = (function (_super) {
    __extends(ExpressionVolatile, _super);
    function ExpressionVolatile() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.lengthMax = 0;
        return _this;
    }
    ExpressionVolatile.prototype.bytesMax = function () {
        return this.lengthMax;
    };
    ExpressionVolatile.prototype.getFixedSizeExpression = function () {
        return this;
    };
    return ExpressionVolatile;
}(ExpressionVariable));
exports.ExpressionVolatile = ExpressionVolatile;
