"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var expression_1 = require("./expression");
var operand_1 = require("./operand");
var util_1 = require("./util");
var Instruction = (function (_super) {
    __extends(Instruction, _super);
    function Instruction() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.mnemonic = null;
        _this.opts = null;
        return _this;
    }
    Instruction.prototype.build = function () {
        _super.prototype.build.call(this);
        return this;
    };
    Instruction.prototype.write = function (arr) { };
    Instruction.prototype.toStringExpression = function () {
        var parts = [];
        parts.push(this.mnemonic.getName());
        if ((parts.join(' ')).length < 8)
            parts.push((new Array(7 - (parts.join(' ')).length)).join(' '));
        if (this.ops)
            if (this.ops.list.length)
                parts.push(this.ops.toString());
        return parts.join(' ');
    };
    Instruction.prototype.toString = function (margin, comment) {
        if (margin === void 0) { margin = '    '; }
        if (comment === void 0) { comment = true; }
        var expression = this.toStringExpression();
        var cmt = '';
        if (comment) {
            var octets = [];
            this.write(octets);
            octets = octets.map(function (byte) {
                return byte <= 0xF ? '0' + byte.toString(16).toUpperCase() : byte.toString(16).toUpperCase();
            });
            cmt = "0x" + octets.join(', 0x') + (" " + this.bytes() + " bytes");
        }
        return this.formatToString(margin, expression, cmt);
    };
    return Instruction;
}(expression_1.ExpressionVolatile));
exports.Instruction = Instruction;
var InstructionSet = (function (_super) {
    __extends(InstructionSet, _super);
    function InstructionSet(ops, matches, opts) {
        var _this = _super.call(this, ops) || this;
        _this.matches = null;
        _this.insn = [];
        _this.picked = -1;
        _this.opts = null;
        _this.matches = matches;
        _this.opts = opts;
        return _this;
    }
    InstructionSet.prototype.write = function (arr) {
        if (this.picked === -1)
            throw Error('Instruction candidates not reduced.');
        this.getPicked().write(arr);
    };
    InstructionSet.prototype.getPicked = function () {
        return this.insn[this.picked];
    };
    InstructionSet.prototype.allInstructionsSameSize = function () {
        var size = this.insn[0].bytes();
        for (var i = 1; i < this.insn.length; i++) {
            if (size !== this.insn[i].bytes())
                return false;
        }
        return true;
    };
    InstructionSet.prototype.getFixedSizeExpression = function () {
        var _a, _b;
        var shortestInd = -1;
        var shortestLen = Infinity;
        for (var m = 0; m < this.ops.list.length; m++) {
            var op = this.ops.list[m];
            if (op instanceof operand_1.Relative) {
                for (var j = 0; j < this.insn.length; j++) {
                    var ins = this.insn[j];
                    var rel = ins.ops.list[m];
                    var success = rel.canHoldMaxOffset(this);
                    if (success) {
                        if (shortestInd === -1) {
                            _a = [j, ins.bytes()], shortestInd = _a[0], shortestLen = _a[1];
                        }
                        else {
                            var bytes = ins.bytes();
                            if (bytes < shortestLen) {
                                _b = [j, bytes], shortestInd = _b[0], shortestLen = _b[1];
                            }
                        }
                    }
                }
            }
        }
        if (shortestInd === -1) {
            var instruction = this.pickShortestInstruction();
            if (!instruction)
                throw Error("Could not fix size for [" + this.index + "] Expression.\n" + this.toString());
            return instruction;
        }
        this.picked = shortestInd;
        return this.getPicked();
    };
    InstructionSet.prototype.evaluate = function () {
        var picked = this.getPicked();
        return picked.evaluate();
    };
    InstructionSet.prototype.bytes = function () {
        return this.picked === -1 ? expression_1.SIZE_UNKNOWN : this.getPicked().bytes();
    };
    InstructionSet.prototype.bytesMax = function () {
        var max = 0;
        for (var _i = 0, _a = this.insn; _i < _a.length; _i++) {
            var ins = _a[_i];
            if (ins) {
                var bytes = ins.bytesMax();
                if (bytes > max)
                    max = bytes;
            }
        }
        return bytes;
    };
    InstructionSet.prototype.calcOffset = function () {
        _super.prototype.calcOffset.call(this);
        var picked = this.getPicked();
        if (picked) {
            picked.offset = this.offset;
        }
    };
    InstructionSet.prototype.pickShortestInstruction = function () {
        if (this.insn.length === 1)
            return this.insn[0];
        if (this.ops.hasRelative())
            return null;
        var size = expression_1.SIZE_UNKNOWN;
        var isize = 0;
        for (var j = 0; j < this.insn.length; j++) {
            var insn = this.insn[j];
            isize = insn.bytes();
            if (isize === expression_1.SIZE_UNKNOWN) {
                this.picked = -1;
                return null;
            }
            if ((size === expression_1.SIZE_UNKNOWN) || (isize < size)) {
                size = isize;
                this.picked = j;
            }
        }
        return this.getPicked();
    };
    InstructionSet.prototype.cloneOperands = function () {
        return this.ops.clone(operand_1.Operands);
    };
    InstructionSet.prototype.createInstructionOperands = function (insn, tpls) {
        var ops = this.cloneOperands();
        for (var j = 0; j < ops.list.length; j++) {
            var op = ops.list[j];
            if (op instanceof operand_1.Operand) {
                if (op instanceof operand_1.Relative) {
                    var Clazz = tpls[j];
                    if (Clazz.name.indexOf('Relative') === 0) {
                        var RelativeClass = Clazz;
                        var rel = op.clone();
                        rel.cast(RelativeClass);
                        ops.list[j] = rel;
                    }
                }
            }
            else if (operand_1.isTnumber(op)) {
                var tpl = tpls[j];
                var num = op;
                if (typeof tpl === 'number') {
                    ops.list[j] = null;
                }
                else if (typeof tpl === 'function') {
                    var Clazz = tpl;
                    if (Clazz.name.indexOf('Relative') === 0) {
                        var RelativeClass = Clazz;
                        var rel = new operand_1.Relative(insn, num);
                        rel.cast(RelativeClass);
                        ops.list[j] = rel;
                    }
                    else if (Clazz.name.indexOf('Immediate') === 0) {
                        var ImmediateClass = Clazz;
                        var imm = new ImmediateClass(num);
                        ops.list[j] = imm;
                    }
                    else
                        throw TypeError('Invalid definition expected Immediate or Relative.');
                }
                else
                    throw TypeError('Invalid definition expected Immediate or Relative or number.');
            }
            else
                throw TypeError('Invalid operand expected Register, Memory, Relative, number or number64.');
        }
        return ops;
    };
    InstructionSet.prototype.build = function () {
        _super.prototype.build.call(this);
        var matches = this.matches;
        var len = matches.length;
        this.insn = [];
        for (var j = 0; j < len; j++) {
            var match = matches[j];
            var insn = this.asm.instruction();
            insn.index = this.index;
            insn.mnemonic = match.mnemonic;
            insn.opts = this.opts;
            var ops = this.createInstructionOperands(insn, match.operandMatch);
            ops.validateSize();
            insn.ops = ops;
            insn.asm = this.asm;
            insn.build();
            this.insn.push(insn);
        }
    };
    InstructionSet.prototype.toString = function (margin, comment) {
        if (margin === void 0) { margin = '    '; }
        if (comment === void 0) { comment = true; }
        if (this.picked === -1) {
            var expression = '(one of:)';
            var spaces = util_1.repeat(' ', Math.max(0, expression_1.Expression.commentColls - expression.length));
            expression += spaces + ("; " + this.formatOffset() + " max " + this.bytesMax() + " bytes\n");
            var lines = [];
            for (var _i = 0, _a = this.matches; _i < _a.length; _i++) {
                var match = _a[_i];
                lines.push(margin + match.mnemonic.toString());
            }
            return expression + lines.join('\n');
        }
        else {
            var picked = this.getPicked();
            return picked.toString(margin, comment) + ' ' + picked.bytes() + ' bytes';
        }
    };
    return InstructionSet;
}(expression_1.ExpressionVolatile));
exports.InstructionSet = InstructionSet;
