"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var expression_1 = require("../../expression");
var operand_1 = require("../../operand");
var util_1 = require("../../util");
var formatOctets_1 = require("./formatOctets");
var Data = (function (_super) {
    __extends(Data, _super);
    function Data(octets) {
        if (octets === void 0) { octets = []; }
        var _this = _super.call(this) || this;
        _this.octets = octets;
        _this.length = octets.length;
        return _this;
    }
    Data.numbersToOctets = function (numbers, size, littleEndian) {
        if (littleEndian === void 0) { littleEndian = true; }
        if (size === operand_1.SIZE.Q)
            return Data.quadsToOctets(numbers, littleEndian);
        var num = numbers;
        var octets = new Array(numbers.length * size);
        if (littleEndian)
            for (var i = 0; i < numbers.length; i++)
                for (var j = 0; j < size; j++)
                    octets[i * size + j] = (num[i] >> (j * 8)) & 0xFF;
        else
            for (var i = 0; i < numbers.length; i++)
                for (var j = 0; j < size; j++)
                    octets[i * size + j] = (num[i] >> ((size - j - 1) * 8)) & 0xFF;
        return octets;
    };
    Data.wordsToOctets = function (words, littleEndian) {
        if (littleEndian === void 0) { littleEndian = true; }
        return Data.numbersToOctets(words, 2, littleEndian);
    };
    Data.doublesToOctets = function (doubles, littleEndian) {
        if (littleEndian === void 0) { littleEndian = true; }
        return Data.numbersToOctets(doubles, 4, littleEndian);
    };
    Data.quadsToOctets = function (quads, littleEndian) {
        if (littleEndian === void 0) { littleEndian = true; }
        if (!(quads instanceof Array))
            throw TypeError('Quads must be and array of number or [number, number].');
        if (!quads.length)
            return [];
        var doubles = new Array(quads.length * 2);
        if (typeof quads[0] === 'number') {
            var qnumbers = quads;
            for (var i = 0; i < qnumbers.length; i++) {
                var hi = util_1.UInt64.hi(qnumbers[i]);
                var lo = util_1.UInt64.lo(qnumbers[i]);
                if (littleEndian) {
                    doubles[i * 2 + 0] = lo;
                    doubles[i * 2 + 1] = hi;
                }
                else {
                    doubles[i * 2 + 0] = hi;
                    doubles[i * 2 + 1] = lo;
                }
            }
        }
        else if (quads[0] instanceof Array) {
            var numbers64 = quads;
            for (var i = 0; i < numbers64.length; i++) {
                var _a = numbers64[i], lo = _a[0], hi = _a[1];
                if (littleEndian) {
                    doubles[i * 2 + 0] = lo;
                    doubles[i * 2 + 1] = hi;
                }
                else {
                    doubles[i * 2 + 0] = hi;
                    doubles[i * 2 + 1] = lo;
                }
            }
        }
        else
            throw TypeError('Quads must be and array of number[] or [number, number][].');
        return Data.doublesToOctets(doubles, littleEndian);
    };
    Data.prototype.write = function (buf) {
        buf.pushArray(this.octets);
    };
    Data.prototype.bytes = function () {
        return this.octets.length;
    };
    Data.prototype.toString = function (margin, comment) {
        if (margin === void 0) { margin = '    '; }
        if (comment === void 0) { comment = true; }
        var datastr = formatOctets_1.default(this.octets);
        var expression = margin + 'db ' + datastr;
        var cmt = '';
        if (comment) {
            var spaces = (new Array(1 + Math.max(0, expression_1.Expression.commentColls - expression.length))).join(' ');
            cmt = spaces + "; " + this.formatOffset() + " " + this.bytes() + " bytes";
        }
        return expression + cmt;
    };
    return Data;
}(expression_1.Expression));
exports.Data = Data;
exports.default = Data;
