"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Data_1 = require("./Data");
var expression_1 = require("../../expression");
var operand_1 = require("../../operand");
var formatOctet_1 = require("./formatOctet");
var formatOctets_1 = require("./formatOctets");
var DataVariable = (function (_super) {
    __extends(DataVariable, _super);
    function DataVariable(ops, littleEndian) {
        if (littleEndian === void 0) { littleEndian = true; }
        var _this = _super.call(this, ops) || this;
        _this.littleEndian = littleEndian;
        _this.octets = new Array(_this.bytes());
        return _this;
    }
    DataVariable.prototype.build = function () {
    };
    DataVariable.prototype.bytes = function () {
        if (this.ops.size <= operand_1.SIZE.NONE)
            throw Error('Unknown operand size in Data.');
        else
            return this.ops.list.length * (this.ops.size >> 3);
    };
    DataVariable.prototype.evaluate = function () {
        var isize = this.ops.size >> 3;
        var list = this.ops.list;
        for (var j = 0; j < list.length; j++) {
            var op = list[j];
            var num;
            if (op instanceof operand_1.Relative) {
                var rel = op;
                num = rel.evaluate(this);
            }
            else if (operand_1.isTnumber(op)) {
                num = op;
            }
            else
                throw Error('Unknow Data operand.');
            var slice = Data_1.default.numbersToOctets([num], isize, this.littleEndian);
            for (var m = 0; m < isize; m++)
                this.octets[j + m] = slice[m];
        }
        return _super.prototype.evaluate.call(this);
    };
    DataVariable.prototype.write = function (buf) {
        buf.pushArray(this.octets);
    };
    DataVariable.prototype.toString = function (margin, comment) {
        if (margin === void 0) { margin = '    '; }
        if (comment === void 0) { comment = true; }
        var datastr = formatOctets_1.default(this.octets);
        var expression = margin + 'dbv ' + datastr;
        var cmt = '';
        if (comment) {
            var spaces = (new Array(1 + Math.max(0, expression_1.Expression.commentColls - expression.length))).join(' ');
            cmt = spaces + "; " + this.formatOffset() + " " + this.bytes() + " bytes";
            if (this.isEvaluated) {
                cmt += ' ' + this.octets.map(function (octet) {
                    return formatOctet_1.default(octet);
                }).join(', ');
            }
        }
        return expression + cmt;
    };
    return DataVariable;
}(expression_1.ExpressionVariable));
exports.DataVariable = DataVariable;
exports.default = DataVariable;
