"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Plugin_1 = require("../Plugin");
var Data_1 = require("./Data");
var DataVariable_1 = require("./DataVariable");
var DataUninitialized_1 = require("./DataUninitialized");
var expression_1 = require("../../expression");
var operand_1 = require("../../operand");
var util_1 = require("../../util");
var fs = require("fs");
var PluginData = (function (_super) {
    __extends(PluginData, _super);
    function PluginData() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    PluginData.prototype.onAsm = function (asm) {
        var _this = this;
        asm.hooks.command.tap('PluginData', function (name, args) {
            switch (name) {
                case 'dbv': return _this.dbv.apply(_this, args);
                case 'db': return _this.db.apply(_this, args);
                case 'dw': return _this.dw.apply(_this, args);
                case 'dd': return _this.dd.apply(_this, args);
                case 'dq': return _this.dq.apply(_this, args);
                case 'resb': return _this.resb(args[0]);
                case 'resw': return _this.resw(args[0]);
                case 'resd': return _this.resd(args[0]);
                case 'resq': return _this.resq(args[0]);
                case 'rest': return _this.rest(args[0]);
                case 'incbin': return _this.incbin.apply(_this, args);
            }
        });
    };
    PluginData.prototype.dbv = function (a, b, c) {
        var ops;
        var littleEndian = this.asm.opts.littleEndian;
        if (a instanceof expression_1.Expression) {
            var expr = a;
            var size = b;
            if (typeof size !== 'number')
                size = this.asm.opts.operandSize;
            else
                size = size << 3;
            if (typeof c === 'boolean')
                littleEndian = c;
            ops = this.asm.ops([expr.rel()], size);
        }
        else if (a instanceof operand_1.Relative) {
            var rel = a;
            var size = b;
            if (typeof size !== 'number')
                size = this.asm.opts.operandSize;
            else
                size = size << 3;
            if (typeof c === 'boolean')
                littleEndian = c;
            ops = this.asm.ops([rel], size);
        }
        else if (a instanceof operand_1.Operands) {
            ops = a;
            if (typeof c === 'boolean')
                littleEndian = c;
        }
        else
            throw TypeError('Data type not supported for DBV.');
        var data = new DataVariable_1.default(ops, littleEndian);
        this.asm.insert(data);
        return data;
    };
    PluginData.prototype.db = function (a, b, c) {
        var octets;
        if (typeof a === 'number')
            return this.db([a], b);
        else if (Array.isArray(a)) {
            octets = a;
            if (typeof b === 'number') {
                var res = octets;
                for (var i = 1; i < b; i++)
                    res = res.concat(octets);
                octets = res;
            }
        }
        else if (typeof a === 'string') {
            var encoding = typeof b === 'string' ? b : 'ascii';
            var buf = new Buffer(a, encoding);
            octets = Array.prototype.slice.call(buf, 0);
            if (typeof b === 'number')
                return this.db(octets, b);
            else
                return this.db(octets, c);
        }
        else if (Buffer.isBuffer(a)) {
            octets = Array.prototype.slice.call(a, 0);
            return this.db(octets, b);
        }
        else if (a instanceof expression_1.Expression)
            return this.dbv(a, b, c);
        else if (a instanceof operand_1.Relative)
            return this.dbv(a, b, c);
        else if (a instanceof operand_1.Operands)
            return this.dbv(a, b);
        else
            throw TypeError('Data type not supported for DB.');
        var data = new Data_1.default(octets);
        this.asm.insert(data);
        return data;
    };
    PluginData.prototype.dw = function (words, littleEndian) {
        if (littleEndian === void 0) { littleEndian = this.asm.opts.littleEndian; }
        if (typeof words === 'number')
            return this.dw([words], littleEndian);
        return this.db(Data_1.default.numbersToOctets(words, 2, littleEndian));
    };
    PluginData.prototype.dd = function (doubles, littleEndian) {
        if (littleEndian === void 0) { littleEndian = this.asm.opts.littleEndian; }
        if (typeof doubles === 'number')
            return this.dd([doubles], littleEndian);
        return this.db(Data_1.default.numbersToOctets(doubles, 4, littleEndian));
    };
    PluginData.prototype.dq = function (quads, littleEndian) {
        if (littleEndian === void 0) { littleEndian = this.asm.opts.littleEndian; }
        var tnums;
        if (typeof quads === 'number')
            tnums = [quads];
        else
            tnums = quads;
        for (var j = 0; j < tnums.length; j++) {
            var num = tnums[j];
            if (typeof num === 'number')
                tnums[j] = [util_1.UInt64.lo(num), util_1.UInt64.hi(num)];
        }
        return this.db(Data_1.default.quadsToOctets(tnums, littleEndian));
    };
    PluginData.prototype.resb = function (length) {
        var data = new DataUninitialized_1.default(length);
        this.asm.insert(data);
        return data;
    };
    PluginData.prototype.resw = function (length) {
        return this.resb(length * 2);
    };
    PluginData.prototype.resd = function (length) {
        return this.resb(length * 4);
    };
    PluginData.prototype.resq = function (length) {
        return this.resb(length * 8);
    };
    PluginData.prototype.rest = function (length) {
        return this.resb(length * 10);
    };
    PluginData.prototype.incbin = function (filepath, offset, len) {
        if (typeof offset === 'undefined') {
            return this.db(fs.readFileSync(filepath));
        }
        else if (typeof len === 'undefined') {
            if (typeof offset !== 'number')
                throw TypeError('Offset must be a number.');
            var fd = fs.openSync(filepath, 'r');
            var total_len = 0;
            var data = [];
            var CHUNK = 4096;
            var buf = new Buffer(CHUNK);
            var bytes = fs.readSync(fd, buf, 0, CHUNK, offset);
            data.push(buf.slice(0, bytes));
            total_len += bytes;
            while ((bytes > 0) && (total_len < bytes)) {
                buf = new Buffer(4096);
                bytes = fs.readSync(fd, buf, 0, CHUNK, null);
                if (bytes > 0) {
                    data.push(buf.slice(0, bytes));
                    total_len += bytes;
                }
            }
            buf = Buffer.concat(data);
            fs.closeSync(fd);
            return this.db(buf);
        }
        else {
            if (typeof offset !== 'number')
                throw TypeError('Offset must be a number.');
            if (typeof len !== 'number')
                throw TypeError('Length must be a number.');
            var buf = new Buffer(len);
            var fd = fs.openSync(filepath, 'r');
            var bytes = fs.readSync(fd, buf, 0, len, offset);
            buf = buf.slice(0, bytes);
            fs.closeSync(fd);
            return this.db(buf);
        }
    };
    return PluginData;
}(Plugin_1.default));
exports.default = PluginData;
