"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Plugin_1 = require("../Plugin");
var table_1 = require("./table");
var instruction_1 = require("./instruction");
var MnemonicEthereum_1 = require("./MnemonicEthereum");
var operand_1 = require("./operand");
var operand_2 = require("../../operand");
var expression_1 = require("../../expression");
var PluginEthereum = (function (_super) {
    __extends(PluginEthereum, _super);
    function PluginEthereum() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    PluginEthereum.prototype.onAsm = function (asm) {
        var _this = this;
        asm.hooks.command.tap('PluginEthereum', function (name, args) {
            name = name.toUpperCase();
            var opcode = table_1.opcodes[name];
            if (typeof opcode === 'number') {
                if ((name[0] === 'P') && (name.substr(0, 4) === 'PUSH')) {
                    var size = parseInt(name.substr(4));
                    if (Array.isArray(args[0])) {
                        return _this.pushX(args[0], size);
                    }
                    else {
                        return _this.pushX(args, size);
                    }
                }
                else {
                    return _this.mnemonic(name, opcode, Array.isArray(args[0]) ? args[0] : args);
                }
            }
            else if (name === 'PUSH') {
                return _this.push.call(_this, Array.isArray(args[0]) ? args[0] : args);
            }
        });
    };
    PluginEthereum.prototype.mnemonic = function (name, opcode, args) {
        if (args && args.length) {
            if (args[0] instanceof expression_1.Expression) {
                console.log('A', args[0].rel());
            }
            else if (typeof args[0] === 'number') {
                this.push(args);
            }
        }
        var instruction = new instruction_1.InstructionEthereum(new MnemonicEthereum_1.default(name, opcode));
        return this.asm.insert(instruction);
    };
    PluginEthereum.prototype.pushX = function (octets, X) {
        if (X === void 0) { X = 1; }
        if (octets.length !== X)
            throw new Error("PUSH" + X + " command expects " + X + " byte immediate, " + octets.length + " given. Consider using variadic PUSH command instead.");
        var name = 'PUSH' + X;
        var mnemonic = new MnemonicEthereum_1.default(name, table_1.opcodes[name]);
        mnemonic.operandSize = operand_1.SIZE_ETHEREUM['S' + X];
        var bytes = mnemonic.operandSize >> 3;
        var operands = new operand_2.Operands([new operand_1.ConstantEthereum(octets)]);
        var instruction = new instruction_1.InstructionEthereum(mnemonic, operands);
        return this.asm.insert(instruction);
    };
    PluginEthereum.prototype.push = function (octets) {
        var cursor = 0;
        var MAX = 32;
        do {
            var diff = octets.length - cursor;
            if (diff > MAX) {
                this.pushX(octets.slice(cursor, cursor + MAX), MAX);
                cursor += MAX;
            }
            else {
                return this.pushX(octets.slice(cursor), diff);
            }
        } while (true);
    };
    PluginEthereum.prototype.pushRelative = function (relative) {
    };
    return PluginEthereum;
}(Plugin_1.default));
exports.default = PluginEthereum;
