"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("../../operand");
var o = require("../x86/operand");
var instruction_1 = require("../x86/instruction");
var p = require("../x86/parts");
var displacement_1 = require("../x86/operand/displacement");
var register_1 = require("../x86/operand/register");
var InstructionX64 = (function (_super) {
    __extends(InstructionX64, _super);
    function InstructionX64() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    InstructionX64.prototype.needs32To64OperandSizeChange = function () {
        return this.mnemonic.operandSize === operand_1.SIZE.Q;
    };
    InstructionX64.prototype.needsRexPrefix = function () {
        if (this.pfxEx)
            return false;
        if (this.mnemonic.rex)
            return true;
        if (!this.ops.list.length)
            return false;
        if (this.ops.hasExtendedRegister())
            return true;
        var _a = this.ops.list, dst = _a[0], src = _a[1];
        if ((dst === o.sil) || (dst === o.dil) || (dst === o.spl) || (dst === o.bpl) ||
            (src === o.sil) || (src === o.dil) || (src === o.spl) || (src === o.bpl))
            return true;
        if (this.mnemonic.operandSizeDefault === operand_1.SIZE.Q)
            return false;
        if (this.needs32To64OperandSizeChange())
            return true;
        return false;
    };
    InstructionX64.prototype.createPrefixes = function () {
        _super.prototype.createPrefixes.call(this);
        if (this.needsRexPrefix())
            this.createRex();
    };
    InstructionX64.prototype.createRex = function () {
        var _a;
        var _b = this.ops.list, dst = _b[0], src = _b[1];
        if ((dst instanceof register_1.Register8High) || (src instanceof register_1.Register8High))
            throw Error('Cannot encode REX prefix with high 8-bit register.');
        if (this.mnemonic.opEncoding === 'mr')
            _a = [src, dst], dst = _a[0], src = _a[1];
        var W = 0, R = 0, X = 0, B = 0;
        if (this.needs32To64OperandSizeChange() && (this.mnemonic.operandSizeDefault !== operand_1.SIZE.Q))
            W = 1;
        var pos = this.mnemonic.opEncoding.indexOf('m');
        if (pos > -1) {
            var m = this.ops.getMemoryOperand();
            if (m) {
                if (m.base && (m.base.idSize() > 3))
                    B = 1;
                if (m.index && (m.index.idSize() > 3))
                    X = 1;
            }
        }
        if ((dst instanceof register_1.RegisterX86) && (src instanceof register_1.RegisterX86)) {
            if (dst.isExtended())
                R = 1;
            if (src.isExtended())
                B = 1;
        }
        else {
            var r = this.ops.getRegisterOperand();
            var mem = this.ops.getMemoryOperand();
            if (r) {
                if (r.idSize() > 3)
                    if (mem)
                        R = 1;
                    else
                        B = 1;
            }
        }
        this.pfxEx = new p.PrefixRex(W, R, X, B);
        this.length++;
        this.lengthMax++;
    };
    InstructionX64.prototype.createModrm = function () {
        var mem = this.ops.getMemoryOperand();
        if (mem && mem.base && (mem.base instanceof register_1.RegisterRip)) {
            if (mem.index || mem.scale)
                throw TypeError('RIP-relative addressing does not support index and scale addressing.');
            var reg = 0;
            if (this.mnemonic.opreg > -1) {
                reg = this.mnemonic.opreg;
            }
            else {
                var r = this.ops.getRegisterOperand();
                if (r)
                    reg = r.get3bitId();
            }
            this.modrm = new p.Modrm(p.Modrm.MOD.INDIRECT, reg, p.Modrm.RM.INDIRECT_DISP);
            this.length++;
            this.lengthMax++;
        }
        else
            _super.prototype.createModrm.call(this);
    };
    InstructionX64.prototype.fixDisplacementSize = function () {
        var mem = this.ops.getMemoryOperand();
        if (mem && (typeof mem == 'object') && (mem.base instanceof register_1.RegisterRip)) {
        }
        else
            _super.prototype.fixDisplacementSize.call(this);
    };
    InstructionX64.prototype.createDisplacement = function () {
        var mem = this.ops.getMemoryOperand();
        if (mem && (typeof mem == 'object') && (mem.base instanceof register_1.RegisterRip)) {
            if (!mem.displacement)
                mem.disp(0);
            var size = displacement_1.DisplacementValue.SIZE.DISP32;
            if (mem.displacement.size < size)
                mem.displacement.signExtend(size);
            this.displacement = new p.Displacement(mem.displacement);
            this.length += size / 8;
            this.lengthMax += size / 8;
        }
        else
            return _super.prototype.createDisplacement.call(this);
    };
    return InstructionX64;
}(instruction_1.InstructionX86));
exports.InstructionX64 = InstructionX64;
