"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("../../operand");
var consts_1 = require("./consts");
var Mnemonic_1 = require("../../Mnemonic");
var table_1 = require("./table");
var OPS = [];
var Match = (function () {
    function Match(mnemonic, operandMatch) {
        this.mnemonic = mnemonic;
        this.operandMatch = operandMatch;
    }
    return Match;
}());
exports.Match = Match;
var MnemonicX86 = (function (_super) {
    __extends(MnemonicX86, _super);
    function MnemonicX86() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.mnemonic = table_1.defaults.mn;
        _this.operandSize = table_1.defaults.s;
        _this.opcode = table_1.defaults.o;
        _this.operandTemplates = OPS;
        _this.opreg = table_1.defaults.or;
        _this.operandSizeDefault = table_1.defaults.ds;
        _this.lock = table_1.defaults.lock;
        _this.regInOp = table_1.defaults.r;
        _this.opcodeDirectionBit = table_1.defaults.dbit;
        _this.useModrm = table_1.defaults.mr;
        _this.rep = table_1.defaults.rep;
        _this.repne = table_1.defaults.repne;
        _this.prefixes = table_1.defaults.pfx;
        _this.opEncoding = table_1.defaults.en;
        _this.rex = table_1.defaults.rex;
        _this.vex = null;
        _this.evex = null;
        _this.mode = table_1.defaults.mod;
        _this.extensions = table_1.defaults.ext;
        return _this;
    }
    MnemonicX86.prototype.toJson = function () {
        var json = _super.prototype.toJson.call(this);
        if (this.opreg > 0)
            json.opcodeExtensionInModrm = this.opreg;
        if (this.regInOp)
            json.registerInOpcode = true;
        json.operandEncoding = this.opEncoding;
        if (this.lock)
            json.lock = true;
        if (this.opcodeDirectionBit)
            json.setOpcodeDirectionBit = true;
        if (this.vex)
            json.vex = this.vex;
        if (this.evex)
            json.evex = this.evex;
        if (this.prefixes)
            json.extraPrefixes = this.prefixes;
        if (this.rep)
            json.prefixRep = true;
        if (this.repne)
            json.prefixRepne = true;
        if (this.rex)
            json.rex = this.rex;
        if (!this.useModrm)
            json.skipMorm = true;
        json.modes = [];
        if (this.mode) {
            for (var _i = 0, MODES_1 = consts_1.MODES; _i < MODES_1.length; _i++) {
                var mode = MODES_1[_i];
                if (this.mode & consts_1.MODE[mode])
                    json.modes.push(mode);
            }
        }
        if (this.extensions) {
            json.extensions = [];
            for (var _a = 0, _b = this.extensions; _a < _b.length; _a++) {
                var ext = _b[_a];
                json.extensions.push(consts_1.EXT[ext]);
            }
        }
        return json;
    };
    MnemonicX86.prototype.toString = function () {
        var opregstr = '';
        if (this.opreg > -1)
            opregstr = ' /' + this.opreg;
        var lock = this.lock ? ' LOCK' : '';
        var rex = this.rex ? ' REX ' + this.rex : '';
        var vex = this.vex ? ' VEX ' + JSON.stringify(this.vex) : '';
        var evex = this.evex ? ' EVEX ' + JSON.stringify(this.evex) : '';
        var dbit = '';
        if (this.opcodeDirectionBit)
            dbit = ' d-bit';
        return _super.prototype.toString.call(this) + opregstr + lock + rex + vex + evex + dbit;
    };
    MnemonicX86.prototype.matchOperandTemplateCase = function (tpl, operand) {
        if (typeof tpl === 'number') {
            if (tpl === operand)
                return tpl;
            else
                return null;
        }
        else if (typeof tpl === 'object') {
            if (tpl === operand)
                return tpl;
            else
                return null;
        }
        else if (typeof tpl === 'function') {
            if (tpl.prototype instanceof operand_1.Relative) {
                if (operand_1.isTnumber(operand))
                    return tpl;
                else if (operand instanceof operand_1.Relative)
                    return tpl;
                else
                    return null;
            }
            else if (tpl.prototype instanceof operand_1.Immediate) {
                if (!operand_1.isTnumber(operand))
                    return null;
                var ImmediateClass = tpl;
                try {
                    new ImmediateClass(operand);
                    return ImmediateClass;
                }
                catch (e) {
                    return null;
                }
            }
            else {
                if (operand instanceof tpl)
                    return tpl;
                else
                    return null;
            }
        }
        else
            throw TypeError('Invalid operand definition.');
    };
    MnemonicX86.prototype.matchOperandTemplate = function (template, operand) {
        for (var _i = 0, template_1 = template; _i < template_1.length; _i++) {
            var tpl = template_1[_i];
            var match = this.matchOperandTemplateCase(tpl, operand);
            if (match)
                return match;
        }
    };
    MnemonicX86.prototype.match = function (ops, opts) {
        if (opts.size !== operand_1.SIZE.ANY) {
            if (opts.size !== this.operandSize)
                return null;
        }
        if (!this.operandTemplates)
            return null;
        if (this.operandTemplates.length !== ops.list.length)
            return null;
        if (this.vex && ops.has5bitRegister())
            return null;
        if (!ops.list.length)
            return new Match(this, []);
        var matches = [];
        for (var i = 0; i < ops.list.length; i++) {
            var template = this.operandTemplates[i];
            var operand = ops.list[i];
            var match = this.matchOperandTemplate(template, operand);
            if (!match)
                return null;
            matches.push(match);
        }
        return new Match(this, matches);
    };
    return MnemonicX86;
}(Mnemonic_1.default));
exports.default = MnemonicX86;
