"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Plugin_1 = require("../Plugin");
var operand_1 = require("../../operand");
var template_1 = require("./template");
var memory_1 = require("./operand/memory");
var instruction_1 = require("./instruction");
var operand_2 = require("./operand");
var parser_1 = require("./operand/parser");
var operandMap = require("./operand/generator");
var expression_1 = require("../../expression");
var AlignX86_1 = require("./AlignX86");
var PluginX86 = (function (_super) {
    __extends(PluginX86, _super);
    function PluginX86() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    PluginX86.prototype.onAsm = function (asm) {
        var _this = this;
        asm.hooks.command.tap('PluginX86', function (name, args) {
            switch (name) {
                case 'align': return _this.align.apply(_this, args);
                case 'lock': return asm._('tpl', template_1.TemplateX86Lock);
                case 'rex': return asm._('tpl', template_1.TemplateX86Rex, args);
                case 'mem':
                case 'disp':
                    return _this.mem(args[0]);
                case 'imm': return _this.imm.apply(_this, args);
                default: {
                    var operand = operandMap[name];
                    if (operand)
                        return operand;
                }
            }
        });
        asm.hooks.op.tap('PluginX86', parser_1.default);
        asm.hooks.ops.tap('PluginX86', function (operands, size) { return _this.ops(operands, size); });
        asm.hooks.instruction.tap('PluginX86', function () { return new instruction_1.InstructionX86(); });
    };
    PluginX86.prototype.align = function (bytes, fill) {
        if (bytes === void 0) { bytes = 4; }
        if (fill === void 0) { fill = null; }
        var align = new AlignX86_1.default(bytes);
        if (fill !== null) {
            if (typeof fill === 'number')
                align.templates = [[fill]];
            else
                align.templates = fill;
        }
        return this.asm.insert(align);
    };
    PluginX86.prototype.mem = function (disp) {
        if (typeof disp === 'number')
            return memory_1.MemoryX86.factory(this.asm.opts.addressSize).disp(disp);
        else if ((disp instanceof Array) && (disp.length == 2))
            return memory_1.MemoryX86.factory(this.asm.opts.addressSize).disp(disp);
        else
            throw TypeError('Displacement value must be of type number or number64.');
    };
    PluginX86.prototype.imm = function (value, signed) {
        if (signed === void 0) { signed = true; }
        return signed ? new operand_1.Immediate(value) : new operand_1.ImmediateUnsigned(value);
    };
    PluginX86.prototype.ops = function (operands, size) {
        if (Array.isArray(operands)) {
            operands = operands.map(function (op) {
                if (typeof op === 'string') {
                    op = operandMap[op];
                    if (!op)
                        throw new Error("Unknown operand " + JSON.stringify(op) + ".");
                }
                else if (op instanceof expression_1.Expression) {
                    op = op.rel();
                }
                return op;
            });
        }
        return new operand_2.OperandsX86(operands, size);
    };
    return PluginX86;
}(Plugin_1.default));
exports.default = PluginX86;
