"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("../../operand");
var o = require("./operand");
var instruction_1 = require("../../instruction");
var p = require("./parts");
var displacement_1 = require("./operand/displacement");
var register_1 = require("./operand/register");
var Immediate_1 = require("./parts/Immediate");
var InstructionX86 = (function (_super) {
    __extends(InstructionX86, _super);
    function InstructionX86() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.pfxOpSize = null;
        _this.pfxAddrSize = null;
        _this.pfxLock = null;
        _this.pfxRep = null;
        _this.pfxRepne = null;
        _this.pfxSegment = null;
        _this.prefixes = [];
        _this.pfxEx = null;
        _this.opcode = new p.Opcode;
        _this.modrm = null;
        _this.sib = null;
        _this.displacement = null;
        _this.immediates = [];
        _this.regToRegDirectionRegIsDst = false;
        return _this;
    }
    InstructionX86.prototype.build = function () {
        _super.prototype.build.call(this);
        this.pfxOpSize = null;
        this.pfxAddrSize = null;
        this.pfxLock = null;
        this.pfxRep = null;
        this.pfxRepne = null;
        this.pfxSegment = null;
        this.prefixes = [];
        this.pfxEx = null;
        this.opcode = new p.Opcode;
        this.modrm = null;
        this.sib = null;
        this.displacement = null;
        this.immediates = [];
        this.length = 0;
        this.lengthMax = 0;
        this.createPrefixes();
        this.createOpcode();
        this.createModrm();
        this.createSib();
        this.createDisplacement();
        this.createImmediates();
        return this;
    };
    InstructionX86.prototype.writePrefixes = function (arr) {
        if (this.pfxLock)
            this.pfxLock.write(arr);
        if (this.pfxRep)
            this.pfxRep.write(arr);
        if (this.pfxRepne)
            this.pfxRepne.write(arr);
        if (this.pfxAddrSize)
            this.pfxAddrSize.write(arr);
        if (this.pfxSegment)
            this.pfxSegment.write(arr);
        if (this.pfxOpSize)
            this.pfxOpSize.write(arr);
        for (var _i = 0, _a = this.prefixes; _i < _a.length; _i++) {
            var pfx = _a[_i];
            pfx.write(arr);
        }
        if (this.pfxEx)
            this.pfxEx.write(arr);
    };
    InstructionX86.prototype.write = function (arr) {
        this.writePrefixes(arr);
        this.opcode.write(arr);
        if (this.modrm)
            this.modrm.write(arr);
        if (this.sib)
            this.sib.write(arr);
        if (this.displacement)
            this.displacement.write(arr);
        if (this.immediates.length)
            for (var _i = 0, _a = this.immediates; _i < _a.length; _i++) {
                var imm = _a[_i];
                imm.write(arr);
            }
    };
    InstructionX86.prototype.fixDisplacementSize = function () {
        if (this.displacement && this.displacement.value.variable) {
            var variable = this.displacement.value.variable;
            var val = variable.evaluatePreliminary(this);
            var size = operand_1.Constant.sizeClass(val);
            if (size > displacement_1.DisplacementValue.SIZE.DISP8)
                this.length += displacement_1.DisplacementValue.SIZE.DISP32 / 8;
            else
                this.length += displacement_1.DisplacementValue.SIZE.DISP8 / 8;
        }
    };
    InstructionX86.prototype.getFixedSizeExpression = function () {
        this.fixDisplacementSize();
        return _super.prototype.getFixedSizeExpression.call(this);
    };
    InstructionX86.prototype.evaluate = function () {
        this.ops.evaluate(this);
        var max = 2;
        for (var j = 0; j < max; j++) {
            var rel = this.ops.getRelative(j);
            if (rel) {
                var res = rel.result;
                this.immediates[j].value.setValue(res);
            }
        }
        if (this.displacement && this.displacement.value.variable) {
            var value = this.displacement.value;
            var variable = value.variable;
            var val = variable.evaluate(this);
            var size = value.size;
            value.setValue(val);
            if (value.size > size)
                throw Error("Displacement does not fit in " + size + " bits.");
            else
                value.signExtend(size);
        }
        return _super.prototype.evaluate.call(this);
    };
    InstructionX86.prototype.lock = function () {
        if (!this.mnemonic.lock)
            throw Error("Instruction \"" + this.mnemonic.mnemonic + "\" does not support LOCK.");
        this.pfxLock = new p.PrefixLock;
        this.length++;
        this.lengthMax++;
        return this;
    };
    InstructionX86.prototype.bt = function () {
        return this.ds();
    };
    InstructionX86.prototype.bnt = function () {
        return this.cs();
    };
    InstructionX86.prototype.rep = function () {
        if (p.PrefixRep.supported.indexOf(this.mnemonic.mnemonic) === -1)
            throw Error("Instruction \"" + this.mnemonic.mnemonic + "\" does not support REP prefix.");
        this.pfxRep = new p.PrefixRep;
        return this;
    };
    InstructionX86.prototype.repe = function () {
        if (p.PrefixRepe.supported.indexOf(this.mnemonic.mnemonic) === -1)
            throw Error("Instruction \"" + this.mnemonic.mnemonic + "\" does not support REPE/REPZ prefix.");
        this.pfxRep = new p.PrefixRepe;
        return this;
    };
    InstructionX86.prototype.repz = function () {
        return this.repe();
    };
    InstructionX86.prototype.repne = function () {
        if (p.PrefixRepne.supported.indexOf(this.mnemonic.mnemonic) === -1)
            throw Error("Instruction \"" + this.mnemonic.mnemonic + "\" does not support REPNE/REPNZ prefix.");
        this.pfxRepne = new p.PrefixRepne;
        return this;
    };
    InstructionX86.prototype.repnz = function () {
        return this.repne();
    };
    InstructionX86.prototype.cs = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.CS);
        return this;
    };
    InstructionX86.prototype.ss = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.SS);
        return this;
    };
    InstructionX86.prototype.ds = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.DS);
        return this;
    };
    InstructionX86.prototype.es = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.ES);
        return this;
    };
    InstructionX86.prototype.fs = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.FS);
        return this;
    };
    InstructionX86.prototype.gs = function () {
        this.pfxSegment = new p.PrefixStatic(p.PREFIX.GS);
        return this;
    };
    InstructionX86.prototype.toStringExpression = function () {
        var expression = _super.prototype.toStringExpression.call(this);
        if (this.pfxLock)
            expression += " {" + this.pfxLock.toString() + "}";
        if (this.pfxSegment)
            expression += " {" + this.pfxSegment.toString() + "}";
        if (this.opts.mask)
            expression += " {" + this.opts.mask.toString() + "}";
        if (this.opts.z)
            expression += " {z}";
        return expression;
    };
    InstructionX86.prototype.needsOperandSizeOverride = function () {
        if ((this.asm.opts.operandSize === operand_1.SIZE.D) && (this.mnemonic.operandSize === operand_1.SIZE.W))
            return true;
        if ((this.asm.opts.operandSize === operand_1.SIZE.W) && (this.mnemonic.operandSize === operand_1.SIZE.D))
            return true;
        return false;
    };
    InstructionX86.prototype.needsAddressSizeOverride = function () {
        var mem = this.ops.getMemoryOperand();
        if (mem) {
            var reg = mem.reg();
            if (reg && (reg.size !== this.asm.opts.addressSize))
                return true;
        }
        return false;
    };
    InstructionX86.prototype.createPrefixes = function () {
        if (this.needsOperandSizeOverride()) {
            this.pfxOpSize = new p.PrefixOperandSizeOverride;
            this.length++;
            this.lengthMax++;
        }
        if (this.needsAddressSizeOverride()) {
            this.pfxAddrSize = new p.PrefixAddressSizeOverride;
            this.length++;
            this.lengthMax++;
        }
        if (this.mnemonic.vex)
            this.createVexPrefix();
        else if (this.mnemonic.evex)
            this.createEvexPrefix();
        if (this.mnemonic.prefixes) {
            for (var _i = 0, _a = this.mnemonic.prefixes; _i < _a.length; _i++) {
                var val = _a[_i];
                this.prefixes.push(new p.PrefixStatic(val));
            }
            this.length += this.mnemonic.prefixes.length;
            this.lengthMax += this.mnemonic.prefixes.length;
        }
    };
    InstructionX86.prototype.createVexPrefix = function () {
        var R = 1, X = 1, B = 1, vvvv = 15;
        var pos = this.mnemonic.opEncoding.indexOf('v');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (!reg)
                throw Error("Could not find Register operand at position " + pos + " to encode VEX.vvvv");
            vvvv = (~reg.get4bitId()) & 15;
        }
        pos = this.mnemonic.opEncoding.indexOf('r');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (!reg)
                throw Error("Could not find Register operand at position " + pos + " to encode VEX.R");
            if (reg.idSize() > 3)
                R = 0;
        }
        pos = this.mnemonic.opEncoding.indexOf('m');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (reg && (reg.idSize() > 3))
                B = 0;
        }
        var mem = this.ops.getMemoryOperand();
        if (mem) {
            if (mem.base && (mem.base.idSize() > 3))
                B = 0;
            if (mem.index && (mem.index.idSize() > 3))
                X = 0;
        }
        this.pfxEx = new p.PrefixVex(this.mnemonic.vex, R, X, B, vvvv);
        this.length += this.pfxEx.bytes;
        this.lengthMax += this.pfxEx.bytes;
    };
    InstructionX86.prototype.createEvexPrefix = function () {
        var evex = this.pfxEx = new p.PrefixEvex(this.mnemonic.evex);
        this.length += 4;
        this.lengthMax += 4;
        var pos = this.mnemonic.opEncoding.indexOf('v');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (!reg)
                throw Error("Could not find Register operand at position " + pos + " to encode EVEX.vvvv");
            evex.vvvv = (~reg.get4bitId()) & 15;
            evex.Vp = reg.id & 16 ? 0 : 1;
        }
        pos = this.mnemonic.opEncoding.indexOf('r');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (!reg)
                throw Error("Could not find Register operand at position " + pos + " to encode VEX.R");
            var id_size = reg.idSize();
            if (id_size > 3)
                evex.R = 0;
            if (id_size > 4) {
                evex.Rp = 0;
                if (reg.id & 8)
                    evex.R = 0;
                else
                    evex.R = 1;
            }
        }
        pos = this.mnemonic.opEncoding.indexOf('m');
        if (pos > -1) {
            var reg = this.ops.getAtIndexOfClass(pos, register_1.RegisterX86);
            if (reg) {
                if (reg.idSize() > 3)
                    evex.B = 0;
                if (reg.idSize() > 4) {
                    evex.X = 0;
                    if (reg.id & 8)
                        evex.B = 0;
                    else
                        evex.B = 1;
                }
            }
        }
        var mem = this.ops.getMemoryOperand();
        if (mem) {
            if (mem.base && (mem.base.idSize() > 3))
                evex.B = 0;
            if (mem.index && (mem.index.idSize() > 3))
                evex.X = 0;
        }
        if (this.opts.mask)
            this.mask(this.opts.mask);
        if (typeof this.opts.z !== 'undefined')
            this.z(this.opts.z);
    };
    InstructionX86.prototype.mask = function (k) {
        if (!(this.pfxEx instanceof p.PrefixEvex))
            throw Error('Cannot set mask on non-EVEX instruction.');
        if (k.id === 0)
            throw TypeError('Mask register 000 cannot be used as mask.');
        this.pfxEx.aaa = k.get3bitId();
        return this;
    };
    InstructionX86.prototype.z = function (value) {
        if (value === void 0) { value = 1; }
        if (!(this.pfxEx instanceof p.PrefixEvex))
            throw Error('Cannot set z-bit on non-EVEX instruction.');
        this.pfxEx.z = value ? 1 : 0;
        return this;
    };
    InstructionX86.prototype.createOpcode = function () {
        var def = this.mnemonic;
        var opcode = this.opcode;
        opcode.op = def.opcode;
        var _a = this.ops.list, dst = _a[0], src = _a[1];
        if (def.regInOp) {
            if (!dst || (!dst.isRegister()))
                throw TypeError("Operation needs destination Register.");
            opcode.op = (opcode.op & p.Opcode.MASK_OP) | dst.get3bitId();
        }
        else {
            if (this.mnemonic.opcodeDirectionBit) {
                var direction = p.Opcode.DIRECTION.REG_IS_DST;
                if (src instanceof register_1.RegisterX86) {
                    direction = p.Opcode.DIRECTION.REG_IS_SRC;
                }
                if ((dst instanceof register_1.RegisterX86) && (src instanceof register_1.RegisterX86)) {
                    if (this.regToRegDirectionRegIsDst)
                        direction = p.Opcode.DIRECTION.REG_IS_DST;
                    else
                        direction = p.Opcode.DIRECTION.REG_IS_SRC;
                }
                opcode.op = (opcode.op & p.Opcode.MASK_DIRECTION) | direction;
            }
        }
        var bytes = opcode.bytes();
        this.length += bytes;
        this.lengthMax += bytes;
    };
    InstructionX86.prototype.createModrm = function () {
        if (!this.mnemonic.useModrm)
            return;
        if (!this.ops.hasRegisterOrMemory())
            return;
        var encoding = this.mnemonic.opEncoding;
        var mod = 0, reg = 0, rm = 0;
        var _a = this.ops.list, dst = _a[0], src = _a[1];
        var has_opreg = (this.mnemonic.opreg > -1);
        var dst_in_modrm = !this.mnemonic.regInOp && !!dst;
        if (has_opreg || dst_in_modrm) {
            var reg_is_dst = this.mnemonic.opEncoding[0] !== 'm' ? true : false;
            if (has_opreg) {
                reg = this.mnemonic.opreg;
                var r = this.ops.getRegisterOperand();
                if (r) {
                    mod = p.Modrm.MOD.REG_TO_REG;
                    rm = r.get3bitId();
                    this.modrm = new p.Modrm(mod, reg, rm);
                    this.length++;
                    this.lengthMax++;
                    return;
                }
            }
            else {
                if ((encoding.length === 2) && (dst instanceof register_1.RegisterX86) && (src instanceof register_1.RegisterX86)) {
                    mod = p.Modrm.MOD.REG_TO_REG;
                    var regreg = (reg_is_dst ? dst : src);
                    var rmreg = (reg_is_dst ? src : dst);
                    reg = regreg.get3bitId();
                    rm = rmreg.get3bitId();
                    this.modrm = new p.Modrm(mod, reg, rm);
                    this.length++;
                    this.lengthMax++;
                    return;
                }
                var rpos = encoding.indexOf('r');
                var rreg;
                if ((rpos > -1) && (rreg = this.ops.getAtIndexOfClass(rpos, register_1.RegisterX86))) {
                    reg = rreg.get3bitId();
                }
                else {
                    var r = this.ops.getRegisterOperand(this.regToRegDirectionRegIsDst ? 0 : 1);
                    if (!r)
                        r = this.ops.getRegisterOperand();
                    if (r) {
                        mod = p.Modrm.MOD.REG_TO_REG;
                        reg = r.get3bitId();
                    }
                }
            }
            var mpos = encoding.indexOf('m');
            if (mpos > -1) {
                var mreg = this.ops.getAtIndexOfClass(mpos, register_1.RegisterX86);
                if (mreg) {
                    mod = p.Modrm.MOD.REG_TO_REG;
                    rm = mreg.get3bitId();
                    this.modrm = new p.Modrm(mod, reg, rm);
                    this.length++;
                    this.lengthMax++;
                    return;
                }
            }
            else {
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            if (!dst) {
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            var m = this.ops.getMemoryOperand();
            if (!m) {
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            if (!m.base && !m.index && !m.displacement)
                throw TypeError('Invalid Memory reference.');
            if (m.index && !m.scale)
                throw TypeError('Memory Index reference needs Scale factor.');
            if (!m.base && !m.index && m.displacement) {
                m.displacement.signExtend(displacement_1.DisplacementValue.SIZE.DISP32);
                mod = p.Modrm.MOD.INDIRECT;
                rm = p.Modrm.RM.NEEDS_SIB;
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            if (m.base && !m.index) {
                mod = p.Modrm.getModDispSize(m);
                if (mod === p.Modrm.MOD.DISP32)
                    m.displacement.signExtend(displacement_1.DisplacementValue.SIZE.DISP32);
                rm = m.base.get3bitId();
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            if (m.base || m.index) {
                mod = p.Modrm.getModDispSize(m);
                if (m.displacement)
                    if ((mod === p.Modrm.MOD.DISP32) || (mod === p.Modrm.MOD.INDIRECT))
                        m.displacement.signExtend(displacement_1.DisplacementValue.SIZE.DISP32);
                rm = p.Modrm.RM.NEEDS_SIB;
                this.modrm = new p.Modrm(mod, reg, rm);
                this.length++;
                this.lengthMax++;
                return;
            }
            throw Error('Fatal error, unreachable code.');
        }
    };
    InstructionX86.prototype.createSib = function () {
        if (!this.modrm)
            return;
        if (this.modrm.mod === p.Modrm.MOD.REG_TO_REG)
            return;
        if ((this.modrm.rm !== p.Modrm.RM.NEEDS_SIB))
            return;
        var m = this.ops.getMemoryOperand();
        if (!m)
            throw Error('No Memory operand to encode SIB.');
        var scalefactor = 0, I = 0, B = 0;
        if (m.scale)
            scalefactor = m.scale.value;
        if (m.index) {
            I = m.index.get3bitId();
            if (I === p.Sib.INDEX_NONE)
                throw Error("Register " + m.index.toString() + " cannot be used as SIB index.");
        }
        else {
            I = p.Sib.INDEX_NONE;
        }
        if (m.base) {
            B = m.base.get3bitId();
        }
        else
            B = p.Sib.BASE_NONE;
        this.sib = new p.Sib(scalefactor, I, B);
        this.length++;
        this.lengthMax++;
    };
    InstructionX86.prototype.createDisplacement = function () {
        var m = this.ops.getMemoryOperand();
        if (m && m.displacement) {
            this.displacement = new p.Displacement(m.displacement);
            if (m.displacement.variable) {
                this.lengthMax += displacement_1.DisplacementValue.SIZE.DISP32 / 8;
            }
            else {
                var size = this.displacement.value.size / 8;
                this.length += size;
                this.lengthMax += size;
            }
        }
        else if (this.modrm && this.sib && (this.sib.B === p.Sib.BASE_NONE)) {
            var disp = null;
            switch (this.modrm.mod) {
                case p.Modrm.MOD.INDIRECT:
                    disp = new displacement_1.DisplacementValue(0);
                    disp.signExtend(displacement_1.DisplacementValue.SIZE.DISP32);
                    break;
                case p.Modrm.MOD.DISP8:
                    disp = new displacement_1.DisplacementValue(0);
                    disp.signExtend(displacement_1.DisplacementValue.SIZE.DISP8);
                    break;
                case p.Modrm.MOD.DISP32:
                    disp = new displacement_1.DisplacementValue(0);
                    disp.signExtend(displacement_1.DisplacementValue.SIZE.DISP32);
                    break;
            }
            if (disp)
                this.displacement = new p.Displacement(disp);
            var size = this.displacement.value.size / 8;
            this.length += size;
            this.lengthMax += size;
        }
    };
    InstructionX86.prototype.createImmediates = function () {
        var max = 2;
        for (var j = 0; j < max; j++) {
            var imm = this.ops.getImmediate(j);
            var immp = void 0;
            if (imm) {
                immp = new Immediate_1.default(imm);
                this.immediates[j] = immp;
                var size = immp.value.size >> 3;
                this.length += size;
                this.lengthMax += size;
            }
            else {
                var rel = this.ops.getRelative(j);
                if (rel) {
                    var immval = operand_1.Immediate.factory(rel.size, 0);
                    immp = new Immediate_1.default(immval);
                    this.immediates[j] = immp;
                    var size = rel.size >> 3;
                    this.length += size;
                    this.lengthMax += size;
                }
            }
        }
    };
    return InstructionX86;
}(instruction_1.Instruction));
exports.InstructionX86 = InstructionX86;
var InstructionSetX86 = (function (_super) {
    __extends(InstructionSetX86, _super);
    function InstructionSetX86() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    InstructionSetX86.prototype.cloneOperands = function () {
        return this.ops.clone(o.OperandsX86);
    };
    InstructionSetX86.prototype.proxy = function (method, args) {
        if (args === void 0) { args = []; }
        for (var _i = 0, _a = this.insn; _i < _a.length; _i++) {
            var ins = _a[_i];
            ins[method].apply(ins, args);
        }
        return this;
    };
    InstructionSetX86.prototype.lock = function () {
        return this.proxy('lock');
    };
    InstructionSetX86.prototype.bt = function () {
        return this.proxy('bt');
    };
    InstructionSetX86.prototype.bnt = function () {
        return this.proxy('bnt');
    };
    InstructionSetX86.prototype.rep = function () {
        return this.proxy('rep');
    };
    InstructionSetX86.prototype.repe = function () {
        return this.proxy('repe');
    };
    InstructionSetX86.prototype.repz = function () {
        return this.proxy('repz');
    };
    InstructionSetX86.prototype.repnz = function () {
        return this.proxy('repnz');
    };
    InstructionSetX86.prototype.repne = function () {
        return this.proxy('repne');
    };
    InstructionSetX86.prototype.cs = function () {
        return this.proxy('cs');
    };
    InstructionSetX86.prototype.ss = function () {
        return this.proxy('ss');
    };
    InstructionSetX86.prototype.ds = function () {
        return this.proxy('ds');
    };
    InstructionSetX86.prototype.es = function () {
        return this.proxy('es');
    };
    InstructionSetX86.prototype.fs = function () {
        return this.proxy('fs');
    };
    InstructionSetX86.prototype.gs = function () {
        return this.proxy('gs');
    };
    return InstructionSetX86;
}(instruction_1.InstructionSet));
exports.InstructionSetX86 = InstructionSetX86;
