"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var MnemonicX86_1 = require("./MnemonicX86");
var operand_1 = require("../../operand");
var parseVexString_1 = require("./parseVexString");
var parseEvexString_1 = require("./parseEvexString");
var mnemonicFromDefX86 = function (def) {
    var mnemonic = new MnemonicX86_1.default;
    mnemonic.opcode = def.o;
    mnemonic.mnemonic = def.mn;
    mnemonic.operandSize = def.s;
    mnemonic.operandTemplates = [];
    mnemonic.opreg = def.or;
    mnemonic.operandSizeDefault = def.ds;
    mnemonic.lock = def.lock;
    mnemonic.regInOp = def.r;
    mnemonic.opcodeDirectionBit = def.dbit;
    mnemonic.rex = def.rex;
    mnemonic.useModrm = def.mr;
    mnemonic.rep = def.rep;
    mnemonic.repne = def.repne;
    mnemonic.prefixes = def.pfx;
    mnemonic.opEncoding = def.en;
    mnemonic.mode = def.mod;
    mnemonic.extensions = def.ext;
    if (def.vex) {
        if (typeof def.vex === 'string') {
            mnemonic.vex = parseVexString_1.default(def.vex);
        }
        else {
            mnemonic.vex = def.vex;
        }
    }
    if (def.evex) {
        if (typeof def.evex === 'string') {
            mnemonic.evex = parseEvexString_1.default(def.evex);
        }
        else {
            mnemonic.evex = def.evex;
        }
    }
    if (def.ops && def.ops.length) {
        var impliedSize_1 = operand_1.SIZE.NONE;
        for (var _i = 0, _a = def.ops; _i < _a.length; _i++) {
            var operand = _a[_i];
            if (!(operand instanceof Array))
                operand = [operand];
            var flattened = operand.reduce(function (a, b) {
                var cur_size = operand_1.SIZE.NONE;
                if (b instanceof operand_1.Register) {
                    cur_size = b.size;
                }
                else if ((typeof b === 'function') && (b.name.indexOf('Register') === 0)) {
                    cur_size = (new b).size;
                }
                if (cur_size !== operand_1.SIZE.NONE) {
                    if (mnemonic.operandSize <= operand_1.SIZE.NONE) {
                        if (impliedSize_1 > operand_1.SIZE.NONE) {
                            if (impliedSize_1 !== cur_size)
                                throw TypeError('Instruction operand size definition mismatch: ' + mnemonic.mnemonic);
                        }
                        else
                            impliedSize_1 = cur_size;
                    }
                }
                return a.concat(b);
            }, []);
            operand = flattened;
            mnemonic.operandTemplates.push(operand);
        }
        if (mnemonic.operandSize <= operand_1.SIZE.NONE) {
            mnemonic.operandSize = impliedSize_1;
        }
    }
    return mnemonic;
};
exports.default = mnemonicFromDefX86;
