"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var regfile_1 = require("../regfile");
var register_1 = require("./register");
function validateRegId(id, min, max, Clazz) {
    if (typeof id !== 'number')
        throw TypeError(Clazz.name + ' register ID must be a number.');
    if (id < min)
        throw TypeError(Clazz.name + " register ID must be at least " + min + ".");
    if (id > max)
        throw TypeError(Clazz.name + " register ID must be at most " + max + ".");
}
function createRegisterGenerator(Klass, min_id, max_id) {
    if (min_id === void 0) { min_id = 0; }
    if (max_id === void 0) { max_id = 15; }
    var cache;
    return function (id) {
        validateRegId(id, min_id, max_id, Klass);
        if (!cache)
            cache = new Array(max_id + 1);
        if (!cache[id])
            cache[id] = new Klass(id);
        return cache[id];
    };
}
exports.rb = createRegisterGenerator(register_1.Register8, 0, 15);
exports.rw = createRegisterGenerator(register_1.Register16, 0, 15);
exports.rd = createRegisterGenerator(register_1.Register32, 0, 15);
exports.rq = createRegisterGenerator(register_1.Register64, 0, 15);
exports.seg = createRegisterGenerator(register_1.RegisterSegment, 0, 15);
exports.mm = createRegisterGenerator(register_1.RegisterMm, 0, 15);
exports.st = createRegisterGenerator(register_1.RegisterSt, 0, 7);
exports.xmm = createRegisterGenerator(register_1.RegisterXmm, 0, 31);
exports.ymm = createRegisterGenerator(register_1.RegisterYmm, 0, 31);
exports.zmm = createRegisterGenerator(register_1.RegisterZmm, 0, 31);
exports.k = createRegisterGenerator(register_1.RegisterK, 0, 7);
exports.bnd = createRegisterGenerator(register_1.RegisterBounds, 0, 3);
exports.cr = createRegisterGenerator(register_1.RegisterCr, 0, 15);
exports.dr = createRegisterGenerator(register_1.RegisterDr, 0, 15);
exports.st0 = exports.st(0);
exports.al = exports.rb(regfile_1.R8.AL);
exports.bl = exports.rb(regfile_1.R8.BL);
exports.cl = exports.rb(regfile_1.R8.CL);
exports.dl = exports.rb(regfile_1.R8.DL);
exports.sil = exports.rb(regfile_1.R8.SIL);
exports.dil = exports.rb(regfile_1.R8.DIL);
exports.bpl = exports.rb(regfile_1.R8.BPL);
exports.spl = exports.rb(regfile_1.R8.SPL);
exports.r8b = exports.rb(regfile_1.R8.R8B);
exports.r9b = exports.rb(regfile_1.R8.R9B);
exports.r10b = exports.rb(regfile_1.R8.R10B);
exports.r11b = exports.rb(regfile_1.R8.R11B);
exports.r12b = exports.rb(regfile_1.R8.R12B);
exports.r13b = exports.rb(regfile_1.R8.R13B);
exports.r14b = exports.rb(regfile_1.R8.R14B);
exports.r15b = exports.rb(regfile_1.R8.R15B);
exports.ah = new register_1.Register8High(regfile_1.R8H.AH);
exports.bh = new register_1.Register8High(regfile_1.R8H.BH);
exports.ch = new register_1.Register8High(regfile_1.R8H.CH);
exports.dh = new register_1.Register8High(regfile_1.R8H.DH);
exports.ax = exports.rw(regfile_1.R16.AX);
exports.bx = exports.rw(regfile_1.R16.BX);
exports.cx = exports.rw(regfile_1.R16.CX);
exports.dx = exports.rw(regfile_1.R16.DX);
exports.si = exports.rw(regfile_1.R16.SI);
exports.di = exports.rw(regfile_1.R16.DI);
exports.bp = exports.rw(regfile_1.R16.BP);
exports.sp = exports.rw(regfile_1.R16.SP);
exports.r8w = exports.rw(regfile_1.R16.R8W);
exports.r9w = exports.rw(regfile_1.R16.R9W);
exports.r10w = exports.rw(regfile_1.R16.R10W);
exports.r11w = exports.rw(regfile_1.R16.R11W);
exports.r12w = exports.rw(regfile_1.R16.R12W);
exports.r13w = exports.rw(regfile_1.R16.R13W);
exports.r14w = exports.rw(regfile_1.R16.R14W);
exports.r15w = exports.rw(regfile_1.R16.R15W);
exports.eax = exports.rd(regfile_1.R32.EAX);
exports.ebx = exports.rd(regfile_1.R32.EBX);
exports.ecx = exports.rd(regfile_1.R32.ECX);
exports.edx = exports.rd(regfile_1.R32.EDX);
exports.esi = exports.rd(regfile_1.R32.ESI);
exports.edi = exports.rd(regfile_1.R32.EDI);
exports.ebp = exports.rd(regfile_1.R32.EBP);
exports.esp = exports.rd(regfile_1.R32.ESP);
exports.r8d = exports.rd(regfile_1.R32.R8D);
exports.r9d = exports.rd(regfile_1.R32.R9D);
exports.r10d = exports.rd(regfile_1.R32.R10D);
exports.r11d = exports.rd(regfile_1.R32.R11D);
exports.r12d = exports.rd(regfile_1.R32.R12D);
exports.r13d = exports.rd(regfile_1.R32.R13D);
exports.r14d = exports.rd(regfile_1.R32.R14D);
exports.r15d = exports.rd(regfile_1.R32.R15D);
exports.rax = exports.rq(regfile_1.R64.RAX);
exports.rcx = exports.rq(regfile_1.R64.RCX);
exports.rdx = exports.rq(regfile_1.R64.RDX);
exports.rbx = exports.rq(regfile_1.R64.RBX);
exports.rsp = exports.rq(regfile_1.R64.RSP);
exports.rbp = exports.rq(regfile_1.R64.RBP);
exports.rsi = exports.rq(regfile_1.R64.RSI);
exports.rdi = exports.rq(regfile_1.R64.RDI);
exports.r8 = exports.rq(regfile_1.R64.R8);
exports.r9 = exports.rq(regfile_1.R64.R9);
exports.r10 = exports.rq(regfile_1.R64.R10);
exports.r11 = exports.rq(regfile_1.R64.R11);
exports.r12 = exports.rq(regfile_1.R64.R12);
exports.r13 = exports.rq(regfile_1.R64.R13);
exports.r14 = exports.rq(regfile_1.R64.R14);
exports.r15 = exports.rq(regfile_1.R64.R15);
exports.rip = new register_1.RegisterRip;
exports.es = exports.seg(regfile_1.SEG.ES);
exports.cs = exports.seg(regfile_1.SEG.CS);
exports.ss = exports.seg(regfile_1.SEG.SS);
exports.ds = exports.seg(regfile_1.SEG.DS);
exports.fs = exports.seg(regfile_1.SEG.FS);
exports.gs = exports.seg(regfile_1.SEG.GS);
exports.cr0 = exports.cr(0);
exports.cr1 = exports.cr(1);
exports.cr2 = exports.cr(2);
exports.cr3 = exports.cr(3);
exports.cr4 = exports.cr(4);
exports.cr5 = exports.cr(5);
exports.cr6 = exports.cr(6);
exports.cr7 = exports.cr(7);
exports.cr8 = exports.cr(8);
exports.dr0 = exports.dr(0);
exports.dr1 = exports.dr(1);
exports.dr2 = exports.dr(2);
exports.dr3 = exports.dr(3);
exports.dr4 = exports.dr(4);
exports.dr5 = exports.dr(5);
exports.dr6 = exports.dr(6);
exports.dr7 = exports.dr(7);
exports.dr8 = exports.dr(8);
