"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var operand_1 = require("../../../operand");
var register_1 = require("./register");
var displacement_1 = require("./displacement");
var regfile_1 = require("../regfile");
var expression_1 = require("../../../expression");
var Scale = (function (_super) {
    __extends(Scale, _super);
    function Scale(scale) {
        if (scale === void 0) { scale = 1; }
        var _this = _super.call(this) || this;
        if (Scale.VALUES.indexOf(scale) < 0)
            throw TypeError("Scale must be one of [1, 2, 4, 8].");
        _this.value = scale;
        return _this;
    }
    Scale.prototype.toString = function () {
        return '' + this.value;
    };
    Scale.VALUES = [1, 2, 4, 8];
    return Scale;
}(operand_1.Operand));
exports.Scale = Scale;
var MemoryX86 = (function (_super) {
    __extends(MemoryX86, _super);
    function MemoryX86() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.base = null;
        _this.index = null;
        _this.scale = null;
        _this.displacement = null;
        return _this;
    }
    MemoryX86.factory = function (size) {
        switch (size) {
            case operand_1.SIZE.B:
                return new Memory8;
            case operand_1.SIZE.W:
                return new Memory16;
            case operand_1.SIZE.D:
                return new Memory32;
            case operand_1.SIZE.Q:
                return new Memory64;
            default:
                return new MemoryX86;
        }
    };
    MemoryX86.prototype.cast = function (size) {
        var mem = MemoryX86.factory(size);
        mem.base = this.base;
        mem.index = this.index;
        mem.scale = this.scale;
        mem.displacement = this.displacement;
        return mem;
    };
    MemoryX86.prototype.reg = function () {
        if (this.base)
            return this.base;
        if (this.index)
            return this.index;
        return null;
    };
    MemoryX86.prototype.needsSib = function () {
        return !!this.index || !!this.scale;
    };
    MemoryX86.prototype.ref = function (base) {
        if (this.index) {
            if (base.size !== this.index.size)
                throw TypeError('Registers dereferencing memory must be of the same size.');
        }
        var is_ebp = (regfile_1.R64.RBP & 7) === base.get3bitId();
        if (is_ebp && !this.displacement)
            this.displacement = new displacement_1.DisplacementValue(0);
        return _super.prototype.ref.call(this, base);
    };
    MemoryX86.prototype.ind = function (index, scale_factor) {
        if (scale_factor === void 0) { scale_factor = 1; }
        if (this.base) {
            if (this.base instanceof register_1.RegisterRip)
                throw TypeError("Cannot have index in memory reference that bases on " + this.base.toString() + ".");
            if (this.base.size !== index.size)
                throw TypeError('Registers dereferencing memory must be of the same size.');
        }
        if (!(index instanceof register_1.RegisterX86))
            throw TypeError('Index must by of type Register.');
        var esp = (regfile_1.R64.RSP & 7);
        if (index.get3bitId() === esp)
            throw TypeError('%esp, %rsp or other 0b100 registers cannot be used as addressing index.');
        this.index = index;
        this.scale = new Scale(scale_factor);
        return this;
    };
    MemoryX86.prototype.disp = function (value) {
        if (value instanceof expression_1.Expression)
            this.displacement = displacement_1.DisplacementValue.fromExpression(value);
        else
            this.displacement = displacement_1.DisplacementValue.fromVariable(value);
        return this;
    };
    MemoryX86.prototype.toString = function () {
        var parts = [];
        if (this.base)
            parts.push(this.base.toString());
        if (this.index)
            parts.push(this.index.toString() + ' * ' + this.scale.toString());
        if (this.displacement)
            parts.push(this.displacement.toString());
        return "[" + parts.join(' + ') + "]";
    };
    return MemoryX86;
}(operand_1.Memory));
exports.MemoryX86 = MemoryX86;
var Memory8 = (function (_super) {
    __extends(Memory8, _super);
    function Memory8() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.B;
        return _this;
    }
    return Memory8;
}(MemoryX86));
exports.Memory8 = Memory8;
var Memory16 = (function (_super) {
    __extends(Memory16, _super);
    function Memory16() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.W;
        return _this;
    }
    return Memory16;
}(MemoryX86));
exports.Memory16 = Memory16;
var Memory32 = (function (_super) {
    __extends(Memory32, _super);
    function Memory32() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.D;
        return _this;
    }
    return Memory32;
}(MemoryX86));
exports.Memory32 = Memory32;
var Memory64 = (function (_super) {
    __extends(Memory64, _super);
    function Memory64() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.Q;
        return _this;
    }
    return Memory64;
}(MemoryX86));
exports.Memory64 = Memory64;
var Memory128 = (function (_super) {
    __extends(Memory128, _super);
    function Memory128() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.O;
        return _this;
    }
    return Memory128;
}(MemoryX86));
exports.Memory128 = Memory128;
var Memory256 = (function (_super) {
    __extends(Memory256, _super);
    function Memory256() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.H;
        return _this;
    }
    return Memory256;
}(MemoryX86));
exports.Memory256 = Memory256;
var Memory512 = (function (_super) {
    __extends(Memory512, _super);
    function Memory512() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.size = operand_1.SIZE.I;
        return _this;
    }
    return Memory512;
}(MemoryX86));
exports.Memory512 = Memory512;
